#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for FcTP state formulation

Authors: Andrew Lee
"""

import pytest
import re
from sys import modules

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    Expression,
    value,
    Var,
    units as pyunits,
)
from pyomo.util.check_units import check_units_equivalent, assert_units_consistent

# Need define_default_scaling_factors, even though it is not used directly
from idaes.models.properties.modular_properties.state_definitions.FcTP import (
    FcTP,
    define_state,
    set_metadata,
)
from idaes.core import (
    MaterialFlowBasis,
    MaterialBalanceType,
    EnergyBalanceType,
    declare_process_block_class,
)
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterData,
)
from idaes.models.properties.modular_properties.base.tests.dummy_eos import DummyEoS
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog


# Note: The state definition is set by importing functions for the relevant module above
@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    pass


@pytest.mark.unit
def test_set_metadata():
    assert set_metadata(None) is None


class TestInvalidBounds(object):
    @pytest.mark.unit
    def test_bad_name(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(
            components={"c1": {}, "c2": {}, "c3": {}},
            phases={"p1": {"equation_of_state": DummyEoS}},
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            state_bounds={"foo": (None, None, None)},
        )

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "props[1] - found unexpected state_bounds key foo. "
                "Please ensure bounds are provided only for expected state "
                "variables and that you have typed the variable names "
                "correctly."
            ),
        ):
            # Build state block
            m.props = m.params.build_state_block([1], defined_state=False)

    @pytest.mark.unit
    def test_mole_frac(self, caplog):
        m = ConcreteModel()

        caplog.set_level(
            idaeslog.WARNING, logger=("idaes.models.properties.modular_properties.")
        )

        m.params = DummyParameterBlock(
            components={"c1": {}, "c2": {}, "c3": {}},
            phases={"p1": {"equation_of_state": DummyEoS}},
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            state_bounds={"mole_frac_comp": (None, None, None)},
        )

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "props[1] - found unexpected state_bounds key "
                "mole_frac_comp. Please ensure bounds are provided only for "
                "expected state variables and that you have typed the "
                "variable names correctly."
            ),
        ):
            # Build state block
            m.props = m.params.build_state_block([1], defined_state=False)


class Test1PhaseDefinedStateFalseNoBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(
            components={"c1": {}, "c2": {}, "c3": {}},
            phases={"p1": {"equation_of_state": DummyEoS}},
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        )

        # Build state block
        m.props = m.params.build_state_block([1], defined_state=False)

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Expression)

        assert isinstance(frame.props[1].flow_mol_comp, Var)
        assert len(frame.props[1].flow_mol_comp) == 3
        for i in frame.props[1].flow_mol_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].flow_mol_comp[i].value is None
        assert check_units_equivalent(
            frame.props[1].flow_mol_comp, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_comp, None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value is None
        assert check_units_equivalent(frame.props[1].pressure, pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value is None
        assert check_units_equivalent(frame.props[1].temperature, pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 1
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value is None
        assert check_units_equivalent(
            frame.props[1].flow_mol_phase, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 1
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1
        assert check_units_equivalent(frame.props[1].phase_frac, None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 3
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [("p1", "c1"), ("p1", "c2"), ("p1", "c3")]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp, None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].mole_frac_comp_eq, Constraint)
        assert len(frame.props[1].mole_frac_comp_eq) == 3
        for i in frame.props[1].mole_frac_comp_eq:
            assert str(frame.props[1].mole_frac_comp_eq[i].body) == str(
                frame.props[1].flow_mol_comp[i]
                - frame.props[1].mole_frac_comp[i]
                * sum(
                    frame.props[1].flow_mol_comp[j] for j in frame.params.component_list
                )
            )

        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1
        assert str(frame.props[1].total_flow_balance.body) == str(
            frame.props[1].flow_mol_phase[frame.params.phase_list.at(1)]
            - frame.props[1].flow_mol
        )

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for i in frame.props[1].component_flow_balances:
            assert i in frame.props[1].params.component_list
            assert str(frame.props[1].component_flow_balances[i].body) == str(
                frame.props[1].mole_frac_comp[i]
                - frame.props[1].mole_frac_phase_comp[frame.params.phase_list.at(1), i]
            )

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 1
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == str(
                frame.props[1].phase_frac[i]
            )

        assert_units_consistent(frame.props[1])


class Test1PhaseDefinedStateTrueWithBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(
            components={"c1": {}, "c2": {}, "c3": {}},
            phases={"p1": {"equation_of_state": DummyEoS}},
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            state_bounds={
                "flow_mol_comp": (0, 100, 200),
                "temperature": (290, 345, 400),
                "pressure": (100000.0, 300000.0, 500000.0),
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        )

        # Build state block
        m.props = m.params.build_state_block([1], defined_state=True)

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Expression)
        assert value(frame.props[1].flow_mol) == 300

        assert isinstance(frame.props[1].flow_mol_comp, Var)
        assert len(frame.props[1].flow_mol_comp) == 3
        for i in frame.props[1].flow_mol_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].flow_mol_comp[i].value == 100
            assert frame.props[1].flow_mol_comp[i].lb == 0
            assert frame.props[1].flow_mol_comp[i].ub == 200
        assert check_units_equivalent(
            frame.props[1].flow_mol_comp, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_comp, None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value == 3e5
        assert frame.props[1].pressure.lb == 1e5
        assert frame.props[1].pressure.ub == 5e5
        assert check_units_equivalent(frame.props[1].pressure, pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value == 345
        assert frame.props[1].temperature.lb == 290
        assert frame.props[1].temperature.ub == 400
        assert check_units_equivalent(frame.props[1].temperature, pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 1
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value == 100
            assert frame.props[1].flow_mol_phase[i].lb == 0
            assert frame.props[1].flow_mol_phase[i].ub == 200
        assert check_units_equivalent(
            frame.props[1].flow_mol_phase, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 1
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1
        assert check_units_equivalent(frame.props[1].phase_frac, None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 3
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [("p1", "c1"), ("p1", "c2"), ("p1", "c3")]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp, None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].mole_frac_comp_eq, Constraint)
        assert len(frame.props[1].mole_frac_comp_eq) == 3
        for i in frame.props[1].mole_frac_comp_eq:
            assert str(frame.props[1].mole_frac_comp_eq[i].body) == str(
                frame.props[1].flow_mol_comp[i]
                - frame.props[1].mole_frac_comp[i]
                * sum(
                    frame.props[1].flow_mol_comp[j] for j in frame.params.component_list
                )
            )

        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1
        assert str(frame.props[1].total_flow_balance.body) == str(
            frame.props[1].flow_mol_phase[frame.params.phase_list.at(1)]
            - frame.props[1].flow_mol
        )

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for i in frame.props[1].component_flow_balances:
            assert i in frame.props[1].params.component_list
            assert str(frame.props[1].component_flow_balances[i].body) == str(
                frame.props[1].mole_frac_comp[i]
                - frame.props[1].mole_frac_phase_comp[frame.params.phase_list.at(1), i]
            )

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 1
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == str(
                frame.props[1].phase_frac[i]
            )

        assert_units_consistent(frame.props[1])


class Test2PhaseDefinedStateFalseNoBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        # Create a dummy parameter block
        m.params = DummyParameterBlock(
            components={"c1": {}, "c2": {}, "c3": {}},
            phases={
                "p1": {"equation_of_state": DummyEoS},
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        )

        # Build state block
        m.props = m.params.build_state_block([1], defined_state=False)

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Expression)

        assert isinstance(frame.props[1].flow_mol_comp, Var)
        assert len(frame.props[1].flow_mol_comp) == 3
        for i in frame.props[1].flow_mol_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].flow_mol_comp[i].value is None
        assert check_units_equivalent(
            frame.props[1].flow_mol_comp, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_comp, None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value is None
        assert check_units_equivalent(frame.props[1].pressure, pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value is None
        assert check_units_equivalent(frame.props[1].temperature, pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 2
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value is None
        assert check_units_equivalent(
            frame.props[1].flow_mol_phase, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 2
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 0.5
        assert check_units_equivalent(frame.props[1].phase_frac, None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 6
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [
                ("p1", "c1"),
                ("p1", "c2"),
                ("p1", "c3"),
                ("p2", "c1"),
                ("p2", "c2"),
                ("p2", "c3"),
            ]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp, None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].mole_frac_comp_eq, Constraint)
        assert len(frame.props[1].mole_frac_comp_eq) == 3
        for i in frame.props[1].mole_frac_comp_eq:
            assert str(frame.props[1].mole_frac_comp_eq[i].body) == str(
                frame.props[1].flow_mol_comp[i]
                - frame.props[1].mole_frac_comp[i]
                * sum(
                    frame.props[1].flow_mol_comp[j] for j in frame.params.component_list
                )
            )

        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1
        assert str(frame.props[1].total_flow_balance.body) == str(
            sum(
                frame.props[1].flow_mol_phase[p]
                for p in frame.props[1].params.phase_list
            )
            - frame.props[1].flow_mol
        )

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for i in frame.props[1].component_flow_balances:
            assert i in frame.props[1].params.component_list
            assert str(frame.props[1].component_flow_balances[i].body) == str(
                frame.props[1].flow_mol_comp[i]
                - sum(
                    frame.props[1].flow_mol_phase[p]
                    * frame.props[1].mole_frac_phase_comp[p, i]
                    for p in frame.props[1].params.phase_list
                )
            )

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 1
        assert str(frame.props[1].sum_mole_frac.body) == str(
            sum(
                frame.props[1].mole_frac_phase_comp[
                    frame.props[1].params.phase_list.at(1), i
                ]
                for i in frame.props[1].params.component_list
            )
            - sum(
                frame.props[1].mole_frac_phase_comp[
                    frame.props[1].params.phase_list.at(2), i
                ]
                for i in frame.props[1].params.component_list
            )
        )

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 2
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == str(
                frame.props[1].phase_frac[i] * frame.props[1].flow_mol
                - frame.props[1].flow_mol_phase[i]
            )

        assert_units_consistent(frame.props[1])


class Test2PhaseDefinedStateTrueWithBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        # Create a dummy parameter block
        m.params = DummyParameterBlock(
            components={"c1": {}, "c2": {}, "c3": {}},
            phases={
                "p1": {"equation_of_state": DummyEoS},
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            state_bounds={
                "flow_mol_comp": (0, 100, 200),
                "temperature": (290, 345, 400),
                "pressure": (100000.0, 300000.0, 500000.0),
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        )

        # Build state block
        m.props = m.params.build_state_block([1], defined_state=True)

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Expression)
        assert value(frame.props[1].flow_mol) == 300

        assert isinstance(frame.props[1].flow_mol_comp, Var)
        assert len(frame.props[1].flow_mol_comp) == 3
        for i in frame.props[1].flow_mol_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].flow_mol_comp[i].value == 100
            assert frame.props[1].flow_mol_comp[i].lb == 0
            assert frame.props[1].flow_mol_comp[i].ub == 200
        assert check_units_equivalent(
            frame.props[1].flow_mol_comp, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_comp, None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value == 3e5
        assert frame.props[1].pressure.lb == 1e5
        assert frame.props[1].pressure.ub == 5e5
        assert check_units_equivalent(frame.props[1].pressure, pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value == 345
        assert frame.props[1].temperature.lb == 290
        assert frame.props[1].temperature.ub == 400
        assert check_units_equivalent(frame.props[1].temperature, pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 2
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value == 50
            assert frame.props[1].flow_mol_phase[i].lb == 0
            assert frame.props[1].flow_mol_phase[i].ub == 200
        assert check_units_equivalent(
            frame.props[1].flow_mol_phase, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 2
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 0.5
        assert check_units_equivalent(frame.props[1].phase_frac, None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 6
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [
                ("p1", "c1"),
                ("p1", "c2"),
                ("p1", "c3"),
                ("p2", "c1"),
                ("p2", "c2"),
                ("p2", "c3"),
            ]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp, None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].mole_frac_comp_eq, Constraint)
        assert len(frame.props[1].mole_frac_comp_eq) == 3
        for i in frame.props[1].mole_frac_comp_eq:
            assert str(frame.props[1].mole_frac_comp_eq[i].body) == str(
                frame.props[1].flow_mol_comp[i]
                - frame.props[1].mole_frac_comp[i]
                * sum(
                    frame.props[1].flow_mol_comp[j] for j in frame.params.component_list
                )
            )

        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1
        assert str(frame.props[1].total_flow_balance.body) == str(
            sum(
                frame.props[1].flow_mol_phase[p]
                for p in frame.props[1].params.phase_list
            )
            - frame.props[1].flow_mol
        )

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for i in frame.props[1].component_flow_balances:
            assert i in frame.props[1].params.component_list
            assert str(frame.props[1].component_flow_balances[i].body) == str(
                frame.props[1].flow_mol_comp[i]
                - sum(
                    frame.props[1].flow_mol_phase[p]
                    * frame.props[1].mole_frac_phase_comp[p, i]
                    for p in frame.props[1].params.phase_list
                )
            )

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 1
        assert str(frame.props[1].sum_mole_frac.body) == str(
            sum(
                frame.props[1].mole_frac_phase_comp[
                    frame.props[1].params.phase_list.at(1), i
                ]
                for i in frame.props[1].params.component_list
            )
            - sum(
                frame.props[1].mole_frac_phase_comp[
                    frame.props[1].params.phase_list.at(2), i
                ]
                for i in frame.props[1].params.component_list
            )
        )

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 2
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == str(
                frame.props[1].phase_frac[i] * frame.props[1].flow_mol
                - frame.props[1].flow_mol_phase[i]
            )

        assert_units_consistent(frame.props[1])


class Test3PhaseDefinedStateFalseNoBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        # Create a dummy parameter block
        m.params = DummyParameterBlock(
            components={"c1": {}, "c2": {}, "c3": {}},
            phases={
                "p1": {"equation_of_state": DummyEoS},
                "p2": {"equation_of_state": DummyEoS},
                "p3": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        )

        # Build state block
        m.props = m.params.build_state_block([1], defined_state=False)

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Expression)

        assert isinstance(frame.props[1].flow_mol_comp, Var)
        assert len(frame.props[1].flow_mol_comp) == 3
        for i in frame.props[1].flow_mol_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].flow_mol_comp[i].value is None
        assert check_units_equivalent(
            frame.props[1].flow_mol_comp, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_comp, None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value is None
        assert check_units_equivalent(frame.props[1].pressure, pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value is None
        assert check_units_equivalent(frame.props[1].temperature, pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 3
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value is None
        assert check_units_equivalent(
            frame.props[1].flow_mol_phase, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 3
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].phase_frac, None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 9
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [
                ("p1", "c1"),
                ("p1", "c2"),
                ("p1", "c3"),
                ("p2", "c1"),
                ("p2", "c2"),
                ("p2", "c3"),
                ("p3", "c1"),
                ("p3", "c2"),
                ("p3", "c3"),
            ]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp, None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].mole_frac_comp_eq, Constraint)
        assert len(frame.props[1].mole_frac_comp_eq) == 3
        for i in frame.props[1].mole_frac_comp_eq:
            assert str(frame.props[1].mole_frac_comp_eq[i].body) == str(
                frame.props[1].flow_mol_comp[i]
                - frame.props[1].mole_frac_comp[i]
                * sum(
                    frame.props[1].flow_mol_comp[j] for j in frame.params.component_list
                )
            )

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for j in frame.props[1].component_flow_balances:
            assert j in frame.params.component_list
            assert str(frame.props[1].component_flow_balances[j].body) == str(
                frame.props[1].flow_mol_comp[j]
                - sum(
                    frame.props[1].flow_mol_phase[p]
                    * frame.props[1].mole_frac_phase_comp[p, j]
                    for p in frame.props[1].params.phase_list
                )
            )

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 3
        for p in frame.props[1].sum_mole_frac:
            assert p in frame.params.phase_list
            assert str(frame.props[1].sum_mole_frac[p].body) == str(
                sum(
                    frame.props[1].mole_frac_phase_comp[p, i]
                    for i in frame.props[1].params.component_list
                )
            )

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 3
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == str(
                frame.props[1].phase_frac[i] * frame.props[1].flow_mol
                - frame.props[1].flow_mol_phase[i]
            )

        assert_units_consistent(frame.props[1])


class Test3PhaseDefinedStateTrueWithBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        # Create a dummy parameter block
        m.params = DummyParameterBlock(
            components={"c1": {}, "c2": {}, "c3": {}},
            phases={
                "p1": {"equation_of_state": DummyEoS},
                "p2": {"equation_of_state": DummyEoS},
                "p3": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            state_bounds={
                "flow_mol_comp": (0, 100, 200),
                "temperature": (290, 345, 400),
                "pressure": (100000.0, 300000.0, 500000.0),
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        )

        # Build state block
        m.props = m.params.build_state_block([1], defined_state=True)

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Expression)
        assert value(frame.props[1].flow_mol) == 300

        assert isinstance(frame.props[1].flow_mol_comp, Var)
        assert len(frame.props[1].flow_mol_comp) == 3
        for i in frame.props[1].flow_mol_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].flow_mol_comp[i].value == 100
            assert frame.props[1].flow_mol_comp[i].lb == 0
            assert frame.props[1].flow_mol_comp[i].ub == 200
        assert check_units_equivalent(
            frame.props[1].flow_mol_comp, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_comp, None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value == 3e5
        assert frame.props[1].pressure.lb == 1e5
        assert frame.props[1].pressure.ub == 5e5
        assert check_units_equivalent(frame.props[1].pressure, pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value == 345
        assert frame.props[1].temperature.lb == 290
        assert frame.props[1].temperature.ub == 400
        assert check_units_equivalent(frame.props[1].temperature, pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 3
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value == 100 / 3
            assert frame.props[1].flow_mol_phase[i].lb == 0
            assert frame.props[1].flow_mol_phase[i].ub == 200
        assert check_units_equivalent(
            frame.props[1].flow_mol_phase, pyunits.mol / pyunits.s
        )

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 3
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].phase_frac, None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 9
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [
                ("p1", "c1"),
                ("p1", "c2"),
                ("p1", "c3"),
                ("p2", "c1"),
                ("p2", "c2"),
                ("p2", "c3"),
                ("p3", "c1"),
                ("p3", "c2"),
                ("p3", "c3"),
            ]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1 / 3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp, None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].mole_frac_comp_eq, Constraint)
        assert len(frame.props[1].mole_frac_comp_eq) == 3
        for i in frame.props[1].mole_frac_comp_eq:
            assert str(frame.props[1].mole_frac_comp_eq[i].body) == str(
                frame.props[1].flow_mol_comp[i]
                - frame.props[1].mole_frac_comp[i]
                * sum(
                    frame.props[1].flow_mol_comp[j] for j in frame.params.component_list
                )
            )

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for j in frame.props[1].component_flow_balances:
            assert j in frame.params.component_list
            assert str(frame.props[1].component_flow_balances[j].body) == str(
                frame.props[1].flow_mol_comp[j]
                - sum(
                    frame.props[1].flow_mol_phase[p]
                    * frame.props[1].mole_frac_phase_comp[p, j]
                    for p in frame.props[1].params.phase_list
                )
            )

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 3
        for p in frame.props[1].sum_mole_frac:
            assert p in frame.params.phase_list
            assert str(frame.props[1].sum_mole_frac[p].body) == str(
                sum(
                    frame.props[1].mole_frac_phase_comp[p, i]
                    for i in frame.props[1].params.component_list
                )
            )

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 3
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == str(
                frame.props[1].phase_frac[i] * frame.props[1].flow_mol
                - frame.props[1].flow_mol_phase[i]
            )

        assert_units_consistent(frame.props[1])


class TestCommon(object):
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(
            components={"c1": {}, "c2": {}, "c3": {}},
            phases={
                "a": {"equation_of_state": DummyEoS},
                "b": {"equation_of_state": DummyEoS},
            },
            state_definition=FcTP,
            pressure_ref=100000.0,
            temperature_ref=300,
            state_bounds={
                "flow_mol_comp": (0, 0.1, 0.2, pyunits.kmol / pyunits.s),
                "temperature": (522, 621, 720, pyunits.degR),
                "pressure": (1, 3, 5, pyunits.bar),
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        )

        # Build state block
        m.props = m.params.build_state_block([1], defined_state=False)

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_convert_vars(self, frame):
        # Check that all state var values and bounds were converted correctly
        assert isinstance(frame.props[1].flow_mol_comp, Var)
        for i in frame.props[1].flow_mol_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].flow_mol_comp[i].value == 100
            assert frame.props[1].flow_mol_comp[i].lb == 0
            assert frame.props[1].flow_mol_comp[i].ub == 200
        assert check_units_equivalent(
            frame.props[1].flow_mol_comp, pyunits.mol / pyunits.s
        )

        assert frame.props[1].pressure.value == 3e5
        assert frame.props[1].pressure.lb == 1e5
        assert frame.props[1].pressure.ub == 5e5
        assert check_units_equivalent(frame.props[1].pressure, pyunits.Pa)

        assert frame.props[1].temperature.value == 345
        assert frame.props[1].temperature.lb == 290
        assert frame.props[1].temperature.ub == 400
        assert check_units_equivalent(frame.props[1].temperature, pyunits.K)

        # Check supporting variables
        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 2

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 6

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 2

        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 1

        assert not hasattr(frame.props[1], "sum_mole_frac_out")

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 2

    @pytest.mark.unit
    def test_calculate_scaling_factors(self, frame):
        frame.props[1].calculate_scaling_factors()

        assert len(frame.props[1].scaling_factor) == 25
        assert frame.props[1].scaling_factor[frame.props[1].dens_mol_phase["a"]] == 1e-2
        assert frame.props[1].scaling_factor[frame.props[1].dens_mol_phase["b"]] == 1e-2
        assert frame.props[1].scaling_factor[frame.props[1].flow_mol] == 1e-2
        assert frame.props[1].scaling_factor[frame.props[1].flow_mol_phase["a"]] == 1e-2
        assert frame.props[1].scaling_factor[frame.props[1].flow_mol_phase["b"]] == 1e-2
        assert frame.props[1].scaling_factor[frame.props[1].flow_mol_comp["c1"]] == 1e-2
        assert frame.props[1].scaling_factor[frame.props[1].flow_mol_comp["c2"]] == 1e-2
        assert frame.props[1].scaling_factor[frame.props[1].flow_mol_comp["c3"]] == 1e-2
        assert (
            frame.props[1].scaling_factor[frame.props[1].flow_mol_phase_comp["a", "c1"]]
            == 1e-2
        )
        assert (
            frame.props[1].scaling_factor[frame.props[1].flow_mol_phase_comp["a", "c2"]]
            == 1e-2
        )
        assert (
            frame.props[1].scaling_factor[frame.props[1].flow_mol_phase_comp["a", "c3"]]
            == 1e-2
        )
        assert (
            frame.props[1].scaling_factor[frame.props[1].flow_mol_phase_comp["b", "c1"]]
            == 1e-2
        )
        assert (
            frame.props[1].scaling_factor[frame.props[1].flow_mol_phase_comp["b", "c2"]]
            == 1e-2
        )
        assert (
            frame.props[1].scaling_factor[frame.props[1].flow_mol_phase_comp["b", "c3"]]
            == 1e-2
        )
        assert frame.props[1].scaling_factor[frame.props[1].mole_frac_comp["c1"]] == 1e3
        assert frame.props[1].scaling_factor[frame.props[1].mole_frac_comp["c2"]] == 1e3
        assert frame.props[1].scaling_factor[frame.props[1].mole_frac_comp["c3"]] == 1e3
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["a", "c1"]
            ]
            == 1e3
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["a", "c2"]
            ]
            == 1e3
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["a", "c3"]
            ]
            == 1e3
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["b", "c1"]
            ]
            == 1e3
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["b", "c2"]
            ]
            == 1e3
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["b", "c3"]
            ]
            == 1e3
        )
        assert frame.props[1].scaling_factor[frame.props[1].pressure] == 1e-5
        assert frame.props[1].scaling_factor[frame.props[1].temperature] == 1e-2

    # Test General Methods
    @pytest.mark.unit
    def test_get_material_flow_terms(self, frame):
        for p, j in frame.params._phase_component_set:
            assert str(frame.props[1].get_material_flow_terms(p, j)) == str(
                frame.props[1].flow_mol_phase_comp[p, j]
            )

    @pytest.mark.unit
    def test_get_enthalpy_flow_terms(self, frame):
        for p in frame.params.phase_list:
            assert str(frame.props[1].get_enthalpy_flow_terms(p)) == str(
                frame.props[1]._enthalpy_flow_term[p]
            )
            assert str(frame.props[1]._enthalpy_flow_term[p].expr) == str(
                frame.props[1].flow_mol_phase[p] * frame.props[1].enth_mol_phase[p]
            )

    @pytest.mark.unit
    def test_get_material_density_terms(self, frame):
        for p, j in frame.params._phase_component_set:
            assert str(frame.props[1].get_material_density_terms(p, j)) == str(
                frame.props[1]._material_density_term[p, j]
            )
            assert str(frame.props[1]._material_density_term[p, j].expr) == str(
                frame.props[1].dens_mol_phase[p]
                * frame.props[1].mole_frac_phase_comp[p, j]
            )

    @pytest.mark.unit
    def test_get_energy_density_terms(self, frame):
        for p in frame.params.phase_list:
            assert str(frame.props[1].get_energy_density_terms(p)) == str(
                frame.props[1]._energy_density_term[p]
            )
            assert str(frame.props[1]._energy_density_term[p].expr) == str(
                frame.props[1].dens_mol_phase[p]
                * frame.props[1].energy_internal_mol_phase[p]
            )

    @pytest.mark.unit
    def test_default_material_balance_type(self, frame):
        assert (
            frame.props[1].default_material_balance_type()
            == MaterialBalanceType.componentTotal
        )

    @pytest.mark.unit
    def test_default_energy_balance_type(self, frame):
        assert (
            frame.props[1].default_energy_balance_type()
            == EnergyBalanceType.enthalpyTotal
        )

    @pytest.mark.unit
    def test_get_material_flow_basis(self, frame):
        assert frame.props[1].get_material_flow_basis() == MaterialFlowBasis.molar

    @pytest.mark.unit
    def test_define_state_vars(self, frame):
        assert frame.props[1].define_state_vars() == {
            "flow_mol_comp": frame.props[1].flow_mol_comp,
            "temperature": frame.props[1].temperature,
            "pressure": frame.props[1].pressure,
        }

    @pytest.mark.unit
    def test_define_display_vars(self, frame):
        assert frame.props[1].define_display_vars() == {
            "Molar Flowrate": frame.props[1].flow_mol_comp,
            "Temperature": frame.props[1].temperature,
            "Pressure": frame.props[1].pressure,
        }

    @pytest.mark.unit
    def test_cloning(self, frame):
        """
        This function tests modular properties are cloned correctly using pyomo's
        clone method. In particular, it tests that all the methods point to variables
        on the clone, and not on the original model.
        """
        blk = frame.clone()

        # Test get_material_flow_terms method
        for p, j in blk.params._phase_component_set:
            assert (
                blk.props[1].get_material_flow_terms(p, j).parent_block()
                is blk.props[1]
            )

        # Test get_enthalpy_flow_terms method
        for p in blk.params.phase_list:
            assert (
                blk.props[1].get_enthalpy_flow_terms(p).parent_block() is blk.props[1]
            )

        # Test get_material_density_terms
        for p, j in blk.params._phase_component_set:
            assert (
                blk.props[1].get_material_density_terms(p, j).parent_block()
                is blk.props[1]
            )

        # Test get_energy_Density_terms
        for p in blk.params.phase_list:
            assert (
                blk.props[1].get_energy_density_terms(p).parent_block() is blk.props[1]
            )

        for i in blk.props[1].define_state_vars().values():
            assert i.parent_block() is blk.props[1]

        for i in blk.props[1].define_display_vars().values():
            assert i.parent_block() is blk.props[1]



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
IDAES Homotopy meta-solver tests.
"""

__author__ = "Andrew Lee"

import pytest

from pyomo.environ import ConcreteModel, Constraint, Param, TerminationCondition, Var

from idaes.core import FlowsheetBlock
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.solvers import get_solver
from idaes.core.solvers.homotopy import homotopy

# Set module level pyest marker
pytestmark = pytest.mark.solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


@pytest.fixture()
def model():
    m = ConcreteModel()
    m.x = Var(initialize=1)
    m.y = Var(initialize=1)

    m.c = Constraint(expr=m.y == m.x**2)

    m.x.fix(10)

    return m


# -----------------------------------------------------------------------------
# Test argument validation
@pytest.mark.unit
def test_invalid_model(model):
    o = object()

    with pytest.raises(TypeError):
        homotopy(o, [model.x], [20])


@pytest.mark.unit
def test_not_var(model):
    model.p = Param()

    with pytest.raises(TypeError):
        homotopy(model, [model.p], [20])


@pytest.mark.unit
def test_var_not_in_model(model):
    m2 = ConcreteModel()

    with pytest.raises(ConfigurationError):
        homotopy(m2, [model.x], [20])


@pytest.mark.unit
def test_var_not_fixed(model):
    model.x.unfix()

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20])


@pytest.mark.unit
def test_current_value_ub(model):
    model.x.setub(5)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20])


@pytest.mark.unit
def test_target_value_ub(model):
    model.x.setub(12)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20])


@pytest.mark.unit
def test_current_value_lb(model):
    model.x.setlb(12)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20])


@pytest.mark.unit
def test_target_value_lb(model):
    model.x.fix(50)
    model.x.setlb(30)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20])


@pytest.mark.unit
def test_step_init(model):
    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], step_init=0.01)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], step_init=0.9)


@pytest.mark.unit
def test_step_cut(model):
    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], step_cut=0.09)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], step_cut=0.91)


@pytest.mark.unit
def test_step_accel(model):
    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], step_accel=-1)


@pytest.mark.unit
def test_iter_target(model):
    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], iter_target=0)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], iter_target=1.7)


@pytest.mark.unit
def test_max_step(model):
    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], max_step=0.04)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], max_step=1.1)


@pytest.mark.unit
def test_min_step(model):
    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], min_step=0.009)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], min_step=0.11)


@pytest.mark.unit
def test_min_max_step(model):
    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], min_step=0.1, max_step=0.05)


@pytest.mark.unit
def test_init_step_min_max(model):
    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], step_init=0.2, max_step=0.1)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], step_init=0.05, min_step=0.1)


@pytest.mark.unit
def test_max_eval(model):
    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], max_eval=0)

    with pytest.raises(ConfigurationError):
        homotopy(model, [model.x], [20], max_eval=1.7)


# -----------------------------------------------------------------------------
# Test termination conditions
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic(model):
    tc, prog, ni = homotopy(model, [model.x], [20])

    assert model.y.value == 400

    assert tc == TerminationCondition.optimal
    assert prog == 1
    assert ni == 4


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_overshoot(model):
    # Use a big step such that overshoot will occur if not caught
    tc, prog, ni = homotopy(model, [model.x], [20], step_init=0.6)

    assert model.y.value == 400

    assert tc == TerminationCondition.optimal
    assert prog == 1
    assert ni == 2


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_constraint_violation(model):
    # Add a constraint to limit y
    model.c2 = Constraint(expr=model.y <= 300)

    # Try to walk to a value of x that gives an infeasible value of y
    tc, prog, ni = homotopy(model, [model.x], [20])

    assert pytest.approx(model.y.value, 1e-4) == 293.27

    assert tc == TerminationCondition.minStepLength
    assert prog == 0.7125
    assert ni == 12


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_max_iter(model):
    tc, prog, ni = homotopy(model, [model.x], [20], max_eval=2)

    assert pytest.approx(model.y.value, 1e-4) == 182.25

    assert tc == TerminationCondition.maxEvaluations
    assert prog == 0.35
    assert ni == 2


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_infeasible_init(model):
    model.c2 = Constraint(expr=model.y <= 50)

    tc, prog, ni = homotopy(model, [model.x], [20])

    assert tc == TerminationCondition.infeasible
    assert prog == 0
    assert ni == 0


# TODO : need tests for convergence with regularisation
# -----------------------------------------------------------------------------
# Test that parameters have correct effect
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_step_accel(model):
    # With zero acceleration, should take 10 steps
    tc, prog, ni = homotopy(model, [model.x], [20], step_accel=0)

    assert model.y.value == 400

    assert tc == TerminationCondition.optimal
    assert prog == 1
    assert ni == 10


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_step_init(model):
    # With zero acceleration and initial step of 0.05, should take 20 steps
    tc, prog, ni = homotopy(model, [model.x], [20], step_init=0.05, step_accel=0)

    assert model.y.value == 400

    assert tc == TerminationCondition.optimal
    assert prog == 1
    assert ni == 20


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_step_cut(model):
    # Add a constraint to limit y
    model.c2 = Constraint(expr=model.y <= 196)

    # Should take 6 steps
    # 4 steps to reach 14, and 2 to cut back to min_step
    tc, prog, ni = homotopy(
        model,
        [model.x],
        [20],
        step_init=0.1,
        min_step=0.025,
        step_cut=0.25,
        step_accel=0,
    )

    assert model.y.value == 196

    assert tc == TerminationCondition.minStepLength
    assert prog == 0.4
    assert ni == 6


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_step_cut_2(model):
    # Add a constraint to limit y
    model.c2 = Constraint(expr=model.y <= 196)

    # Should take 7 steps
    # 4 steps to reach 14, and 3 to cut back to min_step
    tc, prog, ni = homotopy(
        model,
        [model.x],
        [20],
        step_init=0.1,
        min_step=0.01,
        step_cut=0.25,
        step_accel=0,
    )

    assert model.y.value == 196

    assert tc == TerminationCondition.minStepLength
    assert prog == 0.4
    assert ni == 7


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_iter_target(model):
    # Should take 5 steps
    tc, prog, ni = homotopy(model, [model.x], [20], iter_target=2)

    assert model.y.value == 400

    assert tc == TerminationCondition.optimal
    assert prog == 1
    assert ni == 5


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_basic_max_step(model):
    # With max_step = step_init = 0.1, should take 10 steps
    tc, prog, ni = homotopy(model, [model.x], [20], max_step=0.1)

    assert model.y.value == 400

    assert tc == TerminationCondition.optimal
    assert prog == 1
    assert ni == 10


# -----------------------------------------------------------------------------
# Test a more complex problem
@pytest.fixture()
def model2():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # vapor-liquid (ideal) - FTPz
    m.fs.properties_ideal_vl_FTPz = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="Ideal", state_vars="FTPz"
    )
    m.fs.state_block = m.fs.properties_ideal_vl_FTPz.build_state_block(
        defined_state=True
    )

    m.fs.state_block.flow_mol.fix(1)
    m.fs.state_block.temperature.fix(360)
    m.fs.state_block.pressure.fix(101325)
    m.fs.state_block.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state_block.mole_frac_comp["toluene"].fix(0.5)

    assert degrees_of_freedom(m.fs.state_block) == 0

    m.fs.state_block.initialize()

    return m


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_ideal_prop(model2):
    tc, prog, ni = homotopy(model2, [model2.fs.state_block.temperature], [390])

    assert tc == TerminationCondition.optimal
    assert prog == 1
    assert ni == 5

    # Check for VLE results
    assert model2.fs.state_block.mole_frac_phase_comp[
        "Liq", "benzene"
    ].value == pytest.approx(0.291, abs=1e-3)
    assert model2.fs.state_block.mole_frac_phase_comp[
        "Liq", "toluene"
    ].value == pytest.approx(0.709, abs=1e-3)
    assert model2.fs.state_block.mole_frac_phase_comp[
        "Vap", "benzene"
    ].value == pytest.approx(0.5, abs=1e-5)
    assert model2.fs.state_block.mole_frac_phase_comp[
        "Vap", "toluene"
    ].value == pytest.approx(0.5, abs=1e-5)


# Test max_iter here, as a more complicated model is needed
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_ideal_prop_max_iter(model2):
    tc, prog, ni = homotopy(
        model2,
        [model2.fs.state_block.temperature],
        [390],
        max_solver_iterations=3,
        min_step=0.01,
        step_accel=0,
    )

    assert tc == TerminationCondition.optimal
    assert prog == 1
    assert ni == 19

    # Check for VLE results
    assert model2.fs.state_block.mole_frac_phase_comp[
        "Liq", "benzene"
    ].value == pytest.approx(0.291, abs=1e-3)
    assert model2.fs.state_block.mole_frac_phase_comp[
        "Liq", "toluene"
    ].value == pytest.approx(0.709, abs=1e-3)
    assert model2.fs.state_block.mole_frac_phase_comp[
        "Vap", "benzene"
    ].value == pytest.approx(0.5, abs=1e-5)
    assert model2.fs.state_block.mole_frac_phase_comp[
        "Vap", "toluene"
    ].value == pytest.approx(0.5, abs=1e-5)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Base class for unit models
"""

from pyomo.environ import check_optimal_termination
from pyomo.common.config import ConfigValue
from pyomo.network import Port

from idaes.core.base.process_base import (
    declare_process_block_class,
    ProcessBlockData,
    useDefault,
)
from idaes.core.base.property_base import StateBlock
from idaes.core.base.control_volume_base import (
    ControlVolumeBlockData,
    FlowDirection,
    MaterialBalanceType,
)
from idaes.core.util.exceptions import (
    BurntToast,
    ConfigurationError,
    BalanceTypeNotSupportedError,
    InitializationError,
)
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.logger as idaeslog
from idaes.core.solvers import get_solver
from idaes.core.util.config import DefaultBool
from idaes.core.initialization import SingleControlVolumeUnitInitializer


__author__ = "John Eslick, Qi Chen, Andrew Lee"


__all__ = ["UnitModelBlockData", "UnitModelBlock"]

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("UnitModelBlock")
class UnitModelBlockData(ProcessBlockData):
    """
    This is the class for process unit operations models. These are models that
    would generally appear in a process flowsheet or superstructure.
    """

    # Set default initializer
    default_initializer = SingleControlVolumeUnitInitializer

    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            default=useDefault,
            domain=DefaultBool,
            description="Dynamic model flag",
            doc="""Indicates whether this model will be dynamic or not,
**default** = useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - set as a dynamic model,
**False** - set as a steady-state model.}""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=useDefault,
            domain=DefaultBool,
            description="Holdup construction flag",
            doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - construct holdup terms,
**False** - do not construct holdup terms}""",
        ),
    )

    def build(self):
        """
        General build method for UnitModelBlockData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of unit models.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        super(UnitModelBlockData, self).build()

        # Add a placeholder for initialization order
        self._initialization_order = []

        # Set up dynamic flag and time domain
        self._setup_dynamics()

    def model_check(blk):
        """
        This is a general purpose initialization routine for simple unit
        models. This method assumes a single ControlVolume block called
        controlVolume and tries to call the model_check method of the
        controlVolume block. If an AttributeError is raised, the check is
        passed.

        More complex models should overload this method with a model_check
        suited to the particular application, especially if there are multiple
        ControlVolume blocks present.

        Args:
            None

        Returns:
            None
        """
        # Run control volume block model checks
        try:
            blk.controlVolume.model_check()
        except AttributeError:
            pass

    def add_port(self, name, block, doc=None):
        """
        This is a method to build Port objects in a unit model and
        connect these to a specified StateBlock.

        Keyword Args:
            name : name to use for Port object.
            block : an instance of a StateBlock to use as the source to
                    populate the Port object
            doc : doc string for Port object, default = None.

        Returns:
            A Pyomo Port object and associated components.
        """
        # Create Port
        try:
            port, ref_name_list = block.build_port(doc)
        except AttributeError:
            raise ConfigurationError(
                f"{self.name} block object provided to add_port method is not an "
                f"instance of a StateBlock object (does not have a build_port method)."
            )

        # Add Port and References to unit model
        self.add_component(name, port)
        for ref, cname in ref_name_list:
            ref_name = block.get_port_reference_name(cname, name)
            self.add_component(ref_name, ref)

        return port

    def add_inlet_port(self, name=None, block=None, doc=None):
        """
        This is a method to build inlet Port objects in a unit model and
        connect these to a specified control volume or state block.

        The name and block arguments are optional, but must be used together.
        i.e. either both arguments are provided or neither.

        Keyword Args:
            name : name to use for Port object (default = "inlet").
            block : an instance of a ControlVolume or StateBlock to use as the
                    source to populate the Port object. If a ControlVolume is
                    provided, the method will use the inlet state block as
                    defined by the ControlVolume. If not provided, method will
                    attempt to default to an object named control_volume.
            doc : doc string for Port object (default = "Inlet Port")

        Returns:
            A Pyomo Port object and associated components.
        """
        if block is None:
            # Check that name is None
            if name is not None:
                raise ConfigurationError(
                    "{} add_inlet_port was called without a block argument"
                    " but a name argument was provided. Either both "
                    "a name and a block must be provided or neither.".format(self.name)
                )
            else:
                name = "inlet"
            # Try for default ControlVolume name
            try:
                block = self.control_volume
            except AttributeError:
                raise ConfigurationError(
                    "{} add_inlet_port was called without a block argument"
                    " but no default ControlVolume exists "
                    "(control_volume). Please provide block to which the "
                    "Port should be associated.".format(self.name)
                )
        else:
            # Check that name is not None
            if name is None:
                raise ConfigurationError(
                    "{} add_inlet_port was called with a block argument, "
                    "but a name argument was not provided. Either both "
                    "a name and a block must be provided or neither.".format(self.name)
                )

        if doc is None:
            doc = "Inlet Port"

        # Determine type of source block
        if isinstance(block, ControlVolumeBlockData):
            # Work out if this is a 0D or 1D block
            try:
                # Try 0D first
                sblock = block.properties_in
                port, ref_name_list = sblock.build_port(doc)
            except AttributeError:
                # Otherwise a 1D control volume
                try:
                    sblock = block.properties

                    # Need to determine correct subset of indices to add to Port
                    # Look at private attributes on control volume
                    # pylint: disable-next=protected-access
                    if block._flow_direction == FlowDirection.forward:
                        _idx = block.length_domain.first()
                    # pylint: disable-next=protected-access
                    elif block._flow_direction == FlowDirection.backward:
                        _idx = block.length_domain.last()

                    port, ref_name_list = sblock.build_port(
                        doc, slice_index=(slice(None), _idx)
                    )

                except AttributeError:
                    raise ConfigurationError(
                        f"{self.name} - control volume does not have expected "
                        f"names for StateBlocks. Please check that the control "
                        f"volume was constructed correctly."
                    )
        else:
            # Assume a StateBlock indexed only by time
            sblock = block
            port, ref_name_list = sblock.build_port(doc)

        # Add Port and References to unit model
        self.add_component(name, port)
        for ref, cname in ref_name_list:
            ref_name = sblock.get_port_reference_name(cname, name)
            self.add_component(ref_name, ref)

        return port

    def add_outlet_port(self, name=None, block=None, doc=None):
        """
        This is a method to build outlet Port objects in a unit model and
        connect these to a specified control volume or state block.

        The name and block arguments are optional, but must be used together.
        i.e. either both arguments are provided or neither.

        Keyword Args:
            name : name to use for Port object (default = "outlet").
            block : an instance of a ControlVolume or StateBlock to use as the
                    source to populate the Port object. If a ControlVolume is
                    provided, the method will use the outlet state block as
                    defined by the ControlVolume. If not provided, method will
                    attempt to default to an object named control_volume.
            doc : doc string for Port object (default = "Outlet Port")

        Returns:
            A Pyomo Port object and associated components.
        """
        if block is None:
            # Check that name is None
            if name is not None:
                raise ConfigurationError(
                    "{} add_outlet_port was called without a block "
                    "argument but a name argument was provided. Either "
                    "both a name and a block must be provided or neither.".format(
                        self.name
                    )
                )
            else:
                name = "outlet"
            # Try for default ControlVolume name
            try:
                block = self.control_volume
            except AttributeError:
                raise ConfigurationError(
                    "{} add_outlet_port was called without a block "
                    "argument but no default ControlVolume exists "
                    "(control_volume). Please provide block to which the "
                    "Port should be associated.".format(self.name)
                )
        else:
            # Check that name is not None
            if name is None:
                raise ConfigurationError(
                    "{} add_outlet_port was called with a block argument, "
                    "but a name argument was not provided. Either both "
                    "a name and a block must be provided or neither.".format(self.name)
                )

        if doc is None:
            doc = "Outlet Port"

        # Determine type of source block
        if isinstance(block, ControlVolumeBlockData):
            # Work out if this is a 0D or 1D block
            try:
                # Try 0D first
                sblock = block.properties_out
                port, ref_name_list = sblock.build_port(doc)
            except AttributeError:
                # Otherwise a 1D control volume
                try:
                    sblock = block.properties

                    # Need to determine correct subset of indices to add to Port
                    # Look at private attribute on control volume
                    # pylint: disable-next=protected-access
                    if block._flow_direction == FlowDirection.backward:
                        _idx = block.length_domain.first()
                    # pylint: disable-next=protected-access
                    elif block._flow_direction == FlowDirection.forward:
                        _idx = block.length_domain.last()

                    port, ref_name_list = sblock.build_port(
                        doc, slice_index=(slice(None), _idx)
                    )

                except AttributeError:
                    raise ConfigurationError(
                        f"{self.name} - control volume does not have expected "
                        f"names for StateBlocks. Please check that the control "
                        f"volume was constructed correctly."
                    )
        else:
            # Assume a StateBlock indexed only by time
            sblock = block
            port, ref_name_list = sblock.build_port(doc)

        # Add Port and References to unit moodel
        self.add_component(name, port)
        for ref, cname in ref_name_list:
            ref_name = sblock.get_port_reference_name(cname, name)
            self.add_component(ref_name, ref)

        return port

    def add_state_material_balances(self, balance_type, state_1, state_2):
        """
        Method to add material balances linking two State Blocks in a Unit
        Model. This method is not intended to replace Control Volumes, but
        to automate writing material balances linking isolated State Blocks
        in those models where this is required.

        Args:
            balance_type - a MaterialBalanceType Enum indicating the type
                            of material balances to write
            state_1 - first State Block to be linked by balances
            state_2 - second State Block to be linked by balances

        Returns:
            None
        """
        # Confirm that both state blocks stem from the same parameter block
        if not isinstance(state_1, StateBlock):
            raise ConfigurationError(
                "{} state_1 argument to add_state_material_balances "
                "was not an instance of a State Block.".format(self.name)
            )

        if not isinstance(state_2, StateBlock):
            raise ConfigurationError(
                "{} state_2 argument to add_state_material_balances "
                "was not an instance of a State Block.".format(self.name)
            )

        # Check that no constraint with the same name exists
        # We will only support using this method once per Block
        if hasattr(self, "state_material_balances"):
            raise AttributeError(
                "{} a set of constraints named state_material_balances "
                "already exists in the current UnitModel. To avoid "
                "confusion, add_state_material_balances is only supported "
                "once per UnitModel.".format(self.name)
            )

        # Get a representative time point for testing
        rep_time = self.flowsheet().time.first()
        if state_1[rep_time].params is not state_2[rep_time].params:
            raise ConfigurationError(
                "{} add_state_material_balances method was provided with "
                "State Blocks are not linked to the same "
                "instance of a Physical Parameter Block. This method "
                "only supports linking State Blocks from the same "
                "Physical Parameter Block.".format(self.name)
            )

        if balance_type == MaterialBalanceType.useDefault:
            balance_type = state_1[rep_time].default_material_balance_type()

        phase_list = state_1.phase_list
        component_list = state_1.component_list
        pc_set = state_1.phase_component_set

        if balance_type == MaterialBalanceType.componentPhase:
            # TODO : Should we include an optional phase equilibrium term here
            # to allow for systems where a phase-transition may occur?

            @self.Constraint(
                self.flowsheet().time,
                pc_set,
                doc="State material balances",
            )
            def state_material_balances(b, t, p, j):
                return state_1[t].get_material_flow_terms(p, j) == state_2[
                    t
                ].get_material_flow_terms(p, j)

        elif balance_type == MaterialBalanceType.componentTotal:

            @self.Constraint(
                self.flowsheet().time,
                component_list,
                doc="State material balances",
            )
            def state_material_balances(b, t, j):
                return sum(
                    state_1[t].get_material_flow_terms(p, j)
                    for p in phase_list
                    if (p, j) in pc_set
                ) == sum(
                    state_2[t].get_material_flow_terms(p, j)
                    for p in phase_list
                    if (p, j) in pc_set
                )

        elif balance_type == MaterialBalanceType.total:

            @self.Constraint(
                self.flowsheet().time,
                doc="State material balances",
            )
            def state_material_balances(b, t):
                return sum(
                    state_1[t].get_material_flow_terms(p, j) for p, j in pc_set
                ) == sum(state_2[t].get_material_flow_terms(p, j) for p, j in pc_set)

        elif balance_type == MaterialBalanceType.elementTotal:
            raise BalanceTypeNotSupportedError(
                "{} add_state_material_balances does not support "
                "MaterialBalanceType.elementTotal.".format(self.name)
            )
        elif balance_type == MaterialBalanceType.none:
            raise BalanceTypeNotSupportedError(
                "{} add_state_material_balances does not support "
                "MaterialBalanceType.None.".format(self.name)
            )
        else:
            raise BurntToast(
                "{} add_state_material_balances received an unexpected "
                "argument for balance_type. This should never happen. Please "
                "contact the IDAES developers with this bug.".format(self.name)
            )

    def _get_stream_table_contents(self, time_point=0):
        """
        Assume unit has standard configuration of 1 inlet and 1 outlet.

        Developers should overload this as appropriate.
        """
        try:
            return create_stream_table_dataframe(
                {"Inlet": self.inlet, "Outlet": self.outlet}, time_point=time_point
            )
        except AttributeError:
            raise ConfigurationError(
                f"Unit model {self.name} does not have the standard Port "
                f"names (inlet and outlet). Please contact the unit model "
                f"developer to develop a unit specific stream table."
            )

    def initialize(blk, *args, **kwargs):
        """
        Initialization routine for Unit Model objects and associated
        components.

        This method is intended for initializing IDAES unit models and any
        modeling components attached to them, such as costing blocks. This
        method iterates through all objects in blk._initialization_order and
        deactivates them, followed by calling blk.initialize_build. Finally,
        it iterates through all objects in blk._initialization_order in reverse
        and re-activates these whilst calling the associated initialize method.

        Currently, parsing of arguments to the initialize method of attached
        blocks is hard coded - this will be addressed in a future PR.
        Currently, the method supports the following attached components:

        * UnitModelCostingBlocks

        Args:
            costing_args - dict arguments to be passed to costing block
                           initialize method

        For other arguments, see the initialize_unit method.
        """
        # Get any arguments for costing if provided
        cost_args = kwargs.pop("costing_args", {})

        # Get the costing block if present
        # TODO: Clean up in IDAES v2.0
        init_order = blk._initialization_order
        if hasattr(blk, "costing") and blk.costing not in blk._initialization_order:
            # Fallback for older style costing
            init_order.append(blk.costing)

        # If costing block exists, deactivate
        for c in init_order:
            c.deactivate()

        # Remember to collect flags for fixed vars
        flags = blk.initialize_build(*args, **kwargs)

        # If costing block exists, activate and initialize
        for c in init_order:
            c.activate()

            if hasattr(c, "initialize"):
                c.initialize(**cost_args)

        # Return any flags returned by initialize_build
        return flags

    def initialize_build(
        blk, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        This is a general purpose initialization routine for simple unit
        models. This method assumes a single ControlVolume block called
        controlVolume, and first initializes this and then attempts to solve
        the entire unit.

        More complex models should overload this method with their own
        initialization routines,

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) to provide an initial state for
                           initialization (see documentation of the specific
                           property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default IDAES solver)

        Returns:
            None
        """
        if optarg is None:
            optarg = {}

        # Set solver options
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize control volume block
        flags = blk.control_volume.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = opt.solve(blk, tee=slc.tee)

        init_log.info_high(
            "Initialization Step 2 {}.".format(idaeslog.condition(results))
        )

        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.control_volume.release_state(flags, outlvl)

        if not check_optimal_termination(results):
            raise InitializationError(
                f"{blk.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        init_log.info("Initialization Complete: {}".format(idaeslog.condition(results)))

        return None

    def del_component(self, name_or_object):
        """
        Delete a component from this block. Need to introduce code to handle
        un-registering costing blocks
        """
        obj = self.component(name_or_object)

        # TODO: See if Pyomo can give us a call-back to do this
        try:
            # If this is a costing block, need to unregister it
            obj.del_costing()
        except AttributeError:
            pass

        super().del_component(obj)

    def fix_initialization_states(self):
        """
        Attempts to fix inlet states by iterating over all Ports and looking for "inlet"
        in the name.

        More complex models may need to overload this with a model-specific method.

        Returns:
            None
        """
        for p in self.component_objects(Port, descend_into=False):
            if "inlet" in p.name:
                # Assume name is correct and fix all variables inside
                p.fix()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "Douglas Allan"

import pytest
import numpy as np

import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.models_extra.power_generation.unit_models.soc_submodels as soc
import idaes.models_extra.power_generation.unit_models.soc_submodels.testing as soc_testing
from idaes.core.solvers import get_solver

solver = get_solver("ipopt")


@pytest.fixture
def model():
    time_set = [0]
    zfaces = np.linspace(0, 1, 4).tolist()
    m = soc_testing._cell_flowsheet_model(
        dynamic=False, time_set=time_set, zfaces=zfaces
    )
    iznodes = m.fs.iznodes
    tset = m.fs.config.time
    m.fs.temperature_deviation_x = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.K
    )
    m.fs.heat_flux_x0 = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.W / pyo.units.m**2
    )

    m.fs.contact = soc.SocContactResistor(
        control_volume_zfaces=zfaces,
        length_z=m.fs.length_z,
        length_y=m.fs.length_y,
        current_density=m.fs.current_density,
        temperature_z=m.fs.temperature_z,
        temperature_deviation_x=m.fs.temperature_deviation_x,
        heat_flux_x0=m.fs.heat_flux_x0,
    )
    m.fs.temperature_deviation_x.fix(0)
    m.fs.heat_flux_x0.fix(0)

    m.fs.contact.log_preexponential_factor.fix(pyo.log(0.46e-4))
    m.fs.contact.thermal_exponent_dividend.fix(0)
    m.fs.contact.contact_fraction.fix(1)

    return m


@pytest.fixture
def model2():
    time_set = [0, 1]
    zfaces = np.linspace(0, 1, 8).tolist()
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=time_set, time_units=pyo.units.s)
    m.fs.contact = soc.SocContactResistor(control_volume_zfaces=zfaces)
    m.fs.contact.current_density.fix(0)
    m.fs.contact.temperature_z.fix(0)
    m.fs.contact.temperature_deviation_x.fix(0)
    m.fs.contact.heat_flux_x0.fix(0)

    m.fs.contact.log_preexponential_factor.fix(pyo.log(0.46e-4))
    m.fs.contact.thermal_exponent_dividend.fix(0)
    m.fs.contact.contact_fraction.fix(1)

    return m


@pytest.mark.build
@pytest.mark.unit
def test_build(model):
    contact = model.fs.contact
    nz = len(contact.iznodes)
    nt = len(model.fs.time)
    soc_testing._build_test_utility(
        contact,
        comp_dict={
            pyo.Var: {
                "temperature_z": nz * nt,
                "temperature_deviation_x": nz * nt,
                "heat_flux_x0": nz * nt,
                "current_density": nz * nt,
                "length_z": 1,
                "length_y": 1,
                "heat_flux_x1": nz * nt,
                "log_preexponential_factor": 1,
                "thermal_exponent_dividend": 1,
                "contact_fraction": 1,
            },
            pyo.Constraint: {"heat_flux_x_eqn": nz * nt},
            pyo.Expression: {
                "temperature": nz * nt,
                "contact_resistance": nz * nt,
                "voltage_drop_total": nz * nt,
                "joule_heating_flux": nz * nt,
            },
        },
        references=[
            "temperature_z",
            "temperature_deviation_x",
            "heat_flux_x0",
            "current_density",
            "length_z",
            "length_y",
        ],
    )
    assert degrees_of_freedom(model.fs.contact) == 0


@pytest.mark.component
def test_units(model):
    assert_units_consistent(model)


@pytest.mark.build
@pytest.mark.unit
def test_build2(model2):
    contact = model2.fs.contact
    nz = len(contact.iznodes)
    nt = len(model2.fs.time)
    soc_testing._build_test_utility(
        contact,
        comp_dict={
            pyo.Var: {
                "temperature_z": nz * nt,
                "temperature_deviation_x": nz * nt,
                "heat_flux_x0": nz * nt,
                "current_density": nz * nt,
                "length_z": 1,
                "length_y": 1,
                "heat_flux_x1": nz * nt,
                "log_preexponential_factor": 1,
                "thermal_exponent_dividend": 1,
                "contact_fraction": 1,
            },
            pyo.Constraint: {"heat_flux_x_eqn": nz * nt},
            pyo.Expression: {
                "temperature": nz * nt,
                "contact_resistance": nz * nt,
                "voltage_drop_total": nz * nt,
                "joule_heating_flux": nz * nt,
            },
        },
    )
    assert degrees_of_freedom(model2.fs.contact) == 0


@pytest.mark.component
def test_units_2(model2):
    assert_units_consistent(model2)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialization(model):
    model.fs.contact.initialize_build(optarg={"nlp_scaling_method": "user-scaling"})
    model.fs.heat_flux_x0.unfix()
    model.fs.contact.heat_flux_x1.fix()
    model.fs.contact.initialize_build(
        fix_heat_flux_x0=False, optarg={"nlp_scaling_method": "user-scaling"}
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################


def convert_marginal_costs_to_actual_costs(power_marginal_cost_pairs):
    """
    Convert a list of power and marginal cost pairs to a list of power and actual
    cost pairs.

    Args:
        power_marginal_cost_pairs: a list of power and marginal cost pairs (tuple)

    Returns:
        list: a list of power and actual cost pairs (tuple).
    """

    # sort by power
    power_marginal_cost_pairs.sort(key=lambda pair: pair[0])

    # initialize
    actual_costs = []
    pre_power = 0
    pre_cost = 0

    for power, marginal_cost in power_marginal_cost_pairs:

        delta_p = power - pre_power
        cur_cost = pre_cost + marginal_cost * delta_p
        actual_costs.append((power, cur_cost))

        pre_power = power
        pre_cost += marginal_cost * delta_p

    return actual_costs



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for creating additional state variables for electrolyte systems
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

from pyomo.environ import Constraint, NonNegativeReals, Reference, units as pyunits, Var

from idaes.models.properties.modular_properties.base.generic_property import StateIndex
from idaes.models.properties.modular_properties.base.utility import (
    get_bounds_from_config,
)
from idaes.core.util.exceptions import BurntToast
from idaes.core.util.misc import add_object_reference
from idaes.core.base.components import IonData, ApparentData
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale

# Set up logger
_log = idaeslog.getLogger(__name__)


def define_electrolyte_state(b):
    if b.params.config.state_components == StateIndex.true:
        _true_species_state(b)
    elif b.params.config.state_components == StateIndex.apparent:
        _apparent_species_state(b)
    else:
        raise BurntToast(
            "{} - unrecognized value for state_components "
            "argument - this should never happen. Please "
            "contact the IDAES developers".format(b.name)
        )


def calculate_electrolyte_scaling(b):
    if b.params.config.state_components == StateIndex.true:
        _true_species_scaling(b)
    elif b.params.config.state_components == StateIndex.apparent:
        _apparent_species_scaling(b)
    else:
        raise BurntToast(
            "{} - unrecognized value for state_components "
            "argument - this should never happen. Please "
            "contact the IDAES developers".format(b.name)
        )


def _apparent_species_state(b):
    # Create references to base state vars
    add_object_reference(b, "flow_mol_apparent", b.flow_mol)
    b.flow_mol_phase_apparent = Reference(b.flow_mol_phase)
    b.flow_mol_phase_comp_apparent = Reference(b.flow_mol_phase_comp)
    b.mole_frac_phase_comp_apparent = Reference(b.mole_frac_phase_comp)

    # Get units and bounds for true species state
    units = b.params.get_metadata().derived_units
    f_bounds, f_init = get_bounds_from_config(b, "flow_mol", units.FLOW_MOLE)

    # Create true species state vars
    b.flow_mol_phase_comp_true = Var(
        b.params.true_phase_component_set,
        initialize=f_init,
        domain=NonNegativeReals,
        bounds=f_bounds,
        doc="Phase-component molar flowrates of true species",
        units=units.FLOW_MOLE,
    )

    b.mole_frac_phase_comp_true = Var(
        b.params.true_phase_component_set,
        initialize=1 / len(b.params.true_species_set),
        bounds=(1e-20, 1.001),
        doc="Phase-component molar fractions of true species",
        units=pyunits.dimensionless,
    )

    # Check for inherent reactions and add apparent extent terms if required
    if b.has_inherent_reactions:
        b.apparent_inherent_reaction_extent = Var(
            b.params.inherent_reaction_idx,
            initialize=0,
            units=units.FLOW_MOLE,
            doc="Apparent extent of inherent reactions",
        )

    def appr_to_true_species(b, p, j):
        pobj = b.params.get_phase(p)
        cobj = b.params.get_component(j)
        if pobj.is_aqueous_phase:
            if isinstance(cobj, IonData):
                e = 0
                for a in b.params._apparent_set:
                    aobj = b.params.get_component(a)
                    if j in aobj.config.dissociation_species:
                        e += (
                            aobj.config.dissociation_species[j]
                            * b.flow_mol_phase_comp_apparent[p, a]
                        )
            else:
                e = b.flow_mol_phase_comp_apparent[p, j]

            # Next, check for inherent reactions
            if b.has_inherent_reactions:
                for r in b.params.inherent_reaction_idx:
                    # Get stoichiometric coeffiicient for inherent reactions
                    gamma = b.params.inherent_reaction_stoichiometry[r, p, j]

                    if gamma != 0:
                        e += gamma * b.apparent_inherent_reaction_extent[r]

            return b.flow_mol_phase_comp_true[p, j] == e
        else:
            return (
                b.flow_mol_phase_comp_apparent[p, j] == b.flow_mol_phase_comp_true[p, j]
            )

    b.appr_to_true_species = Constraint(
        b.params.true_phase_component_set,
        rule=appr_to_true_species,
        doc="Apparent to true species conversion",
    )

    def true_species_mole_fractions(b, p, j):
        return (
            b.mole_frac_phase_comp_true[p, j]
            * sum(
                b.flow_mol_phase_comp_true[p, k]
                for k in b.params.true_species_set
                if (p, k) in b.params.true_phase_component_set
            )
            == b.flow_mol_phase_comp_true[p, j]
        )

    b.true_mole_frac_constraint = Constraint(
        b.params.true_phase_component_set,
        rule=true_species_mole_fractions,
        doc="Calculation of true species mole fractions",
    )


def _apparent_species_scaling(b):
    for p, j in b.params.true_phase_component_set:
        sf = iscale.get_scaling_factor(
            b.flow_mol_phase_comp_true[p, j], default=1, warning=True
        )

        iscale.constraint_scaling_transform(
            b.appr_to_true_species[p, j], sf, overwrite=False
        )
        iscale.constraint_scaling_transform(
            b.true_mole_frac_constraint[p, j], sf, overwrite=False
        )


def _true_species_state(b):
    # Create references to base state vars
    add_object_reference(b, "flow_mol_true", b.flow_mol)
    b.flow_mol_phase_true = Reference(b.flow_mol_phase)
    b.flow_mol_phase_comp_true = Reference(b.flow_mol_phase_comp)
    b.mole_frac_phase_comp_true = Reference(b.mole_frac_phase_comp)

    # Get units and bounds for apparent species state
    units = b.params.get_metadata().derived_units
    f_bounds, f_init = get_bounds_from_config(b, "flow_mol", units.FLOW_MOLE)

    # Create apparent species state vars
    b.flow_mol_phase_comp_apparent = Var(
        b.params.apparent_phase_component_set,
        initialize=f_init,
        domain=NonNegativeReals,
        bounds=f_bounds,
        doc="Phase-component molar flowrates of apparent species",
        units=units.FLOW_MOLE,
    )

    b.mole_frac_phase_comp_apparent = Var(
        b.params.apparent_phase_component_set,
        initialize=1 / len(b.params.apparent_species_set),
        bounds=(1e-20, 1.001),
        doc="Phase-component molar fractions of apparent species",
        units=pyunits.dimensionless,
    )

    def true_to_appr_species(b, p, j):
        pobj = b.params.get_phase(p)
        cobj = b.params.get_component(j)
        if pobj.is_aqueous_phase():
            total_charge = sum(
                b.flow_mol_phase_comp_true[p, c]
                * b.params.get_component(c).config.charge
                for c in b.params.cation_set
            )

            if isinstance(cobj, ApparentData):
                # Need to recompose composition
                ions = list(cobj.config.dissociation_species.keys())

                e = (
                    b.flow_mol_phase_comp_true[p, ions[0]]
                    * abs(b.params.get_component(ions[0]).config.charge)
                    * b.flow_mol_phase_comp_true[p, ions[1]]
                    / (total_charge * cobj.config.dissociation_species[ions[1]])
                )
            elif j == "H2O":
                # Special case for water
                # Need to generalise to cover other cases with weak acids
                e = b.flow_mol_phase_comp_true[p, j]

                if "H+" in b.params.cation_set and "OH-" in b.params.anion_set:
                    e += (
                        b.flow_mol_phase_comp_true[p, "H+"]
                        * b.flow_mol_phase_comp_true[p, "OH-"]
                        / total_charge
                    )
                elif "H3O+" in b.params.cation_set and "OH-" in b.params.anion_set:
                    # Assume H3O+
                    e += (
                        2
                        * b.flow_mol_phase_comp_true[p, "H3O+"]
                        * b.flow_mol_phase_comp_true[p, "OH-"]
                        / total_charge
                    )
            else:
                e = b.flow_mol_phase_comp_true[p, j]

            return b.flow_mol_phase_comp_apparent[p, j] == e
        else:
            return (
                b.flow_mol_phase_comp_apparent[p, j] == b.flow_mol_phase_comp_true[p, j]
            )

    b.true_to_appr_species = Constraint(
        b.params.apparent_phase_component_set,
        rule=true_to_appr_species,
        doc="True to apparent species conversion",
    )

    def apparent_species_mole_fractions(b, p, j):
        return (
            b.mole_frac_phase_comp_apparent[p, j]
            * sum(
                b.flow_mol_phase_comp_apparent[p, k]
                for k in b.params.apparent_species_set
                if (p, k) in b.params.apparent_phase_component_set
            )
            == b.flow_mol_phase_comp_apparent[p, j]
        )

    b.appr_mole_frac_constraint = Constraint(
        b.params.apparent_phase_component_set,
        rule=apparent_species_mole_fractions,
        doc="Calculation of apparent species mole fractions",
    )


def _true_species_scaling(b):
    pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring

from pyomo.core.base.componentuid import ComponentUID
from pyomo.core.base.expression import Expression


def get_tracking_cost_from_constant_setpoint(
    variables,
    time,
    setpoint_data,
    weight_data=None,
):
    """
    This function returns a tracking cost expression for the given time-indexed
    variables and associated setpoint data.

    Arguments
    ---------
    variables: list
        List of time-indexed variables to include in the tracking cost
        expression
    time: iterable
        Set by which to index the tracking expression
    setpoint_data: dict
        Maps variable names to setpoint values
    weight_data: dict
        Optional. Maps variable names to tracking cost weights. If not
        provided, weights of one are used.

    Returns
    -------
    Pyomo Expression, indexed by time, containing the sum of weighted
    squared difference between variables and setpoint values.

    """
    variable_names = [
        (
            str(ComponentUID(var))
            if not var.is_reference()
            else str(ComponentUID(var.referent))
        )
        for var in variables
    ]
    if weight_data is None:
        weight_data = {name: 1.0 for name in variable_names}
    for i, name in enumerate(variable_names):
        if name not in setpoint_data:
            raise KeyError(
                "Setpoint data dictionary does not contain a key for variable\n"
                "%s with ComponentUID %s" % (variables[i].name, name)
            )
        if name not in weight_data:
            raise KeyError(
                "Tracking weight dictionary does not contain a key for "
                "variable\n%s with ComponentUID %s" % (variables[i].name, name)
            )

    def tracking_rule(m, t):
        return sum(
            weight_data[name] * (var[t] - setpoint_data[name]) ** 2
            for name, var in zip(variable_names, variables)
        )

    tracking_expr = Expression(time, rule=tracking_rule)
    return tracking_expr



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for process_meta

Author: Andrew Lee
"""
import pytest
import re

from pyomo.environ import ConcreteModel, units
from pyomo.util.check_units import assert_units_equivalent

from idaes.core.base.property_meta import PropertyClassMetadata, UnitSet
from idaes.core.util.exceptions import PropertyPackageError


@pytest.mark.unit
def test_invalid_require_base_quantity():
    with pytest.raises(
        PropertyPackageError,
        match=re.escape("Unrecognized units of measurement for quantity TIME (foo)"),
    ):
        us = UnitSet()
        us.set_units(time="foo")


@pytest.mark.unit
def test_mismatched_length_units():
    with pytest.raises(
        PropertyPackageError,
        match=re.escape(
            "Invalid units of measurement for quantity LENGTH (s). "
            "Please ensure units provided are valid for this quantity and use the Pyomo unit registry."
        ),
    ):
        us = UnitSet()
        us.set_units(length=units.s)


@pytest.mark.unit
def test_mismatched_mass_units():
    with pytest.raises(
        PropertyPackageError,
        match=re.escape(
            "Invalid units of measurement for quantity MASS (m). "
            "Please ensure units provided are valid for this quantity and use the Pyomo unit registry."
        ),
    ):
        us = UnitSet()
        us.set_units(mass=units.m)


@pytest.mark.unit
def test_mismatched_amount_units():
    with pytest.raises(
        PropertyPackageError,
        match=re.escape(
            "Invalid units of measurement for quantity AMOUNT (m). "
            "Please ensure units provided are valid for this quantity and use the Pyomo unit registry."
        ),
    ):
        us = UnitSet()
        us.set_units(amount=units.m)


@pytest.mark.unit
def test_mismatched_temperature_units():
    with pytest.raises(
        PropertyPackageError,
        match=re.escape(
            "Invalid units of measurement for quantity TEMPERATURE (m). "
            "Please ensure units provided are valid for this quantity and use the Pyomo unit registry."
        ),
    ):
        us = UnitSet()
        us.set_units(temperature=units.m)


@pytest.mark.unit
def test_mismatched_current_units():
    with pytest.raises(
        PropertyPackageError,
        match=re.escape(
            "Invalid units of measurement for quantity CURRENT (m). "
            "Please ensure units provided are valid for this quantity and use the Pyomo unit registry."
        ),
    ):
        us = UnitSet()
        us.set_units(current=units.m)


@pytest.mark.unit
def test_mismatched_luminous_intensity_units():
    with pytest.raises(
        PropertyPackageError,
        match=re.escape(
            "Invalid units of measurement for quantity LUMINOUS_INTENSITY (m). "
            "Please ensure units provided are valid for this quantity and use the Pyomo unit registry."
        ),
    ):
        us = UnitSet()
        us.set_units(luminous_intensity=units.m)


@pytest.mark.unit
def test_mismatched_time_units():
    with pytest.raises(
        PropertyPackageError,
        match=re.escape(
            "Invalid units of measurement for quantity TIME (m). "
            "Please ensure units provided are valid for this quantity and use the Pyomo unit registry."
        ),
    ):
        us = UnitSet()
        us.set_units(time=units.m)


@pytest.fixture(scope="module")
def unit_set():
    us = UnitSet()
    us.set_units(
        time=units.s,
        length=units.m,
        mass=units.kg,
        amount=units.mol,
        temperature=units.K,
        current=units.ampere,
        luminous_intensity=units.candela,
    )
    return us


@pytest.mark.unit
def test_time(unit_set):
    assert unit_set.TIME == units.s
    assert unit_set["time"] == units.s


@pytest.mark.unit
def test_length(unit_set):
    assert unit_set.LENGTH == units.m
    assert unit_set["length"] == units.m


@pytest.mark.unit
def test_mass(unit_set):
    assert unit_set.MASS == units.kg
    assert unit_set["mass"] == units.kg


@pytest.mark.unit
def test_amount(unit_set):
    assert unit_set.AMOUNT == units.mol
    assert unit_set["amount"] == units.mol


@pytest.mark.unit
def test_temperature(unit_set):
    assert unit_set.TEMPERATURE == units.K
    assert unit_set["temperature"] == units.K


@pytest.mark.unit
def test_current(unit_set):
    assert unit_set.CURRENT == units.ampere
    assert unit_set["current"] == units.ampere


@pytest.mark.unit
def test_luminous_intensity(unit_set):
    assert unit_set.LUMINOUS_INTENSITY == units.candela
    assert unit_set["luminous_intensity"] == units.candela
    # Also check with space
    assert unit_set["luminous intensity"] == units.candela


# ---------------------------------------------------------------------------------------------
# Test derived units
derived_quantities = {
    "area": units.m**2,
    "volume": units.m**3,
    "volume_mass": units.m**3 / units.kg,
    "volume_mole": units.m**3 / units.mol,
    "flow_mass": units.kg * units.s**-1,
    "flow_mole": units.mol * units.s**-1,
    "flow_vol": units.m**3 * units.s**-1,
    "flux_mass": (units.kg * units.s**-1 * units.m**-2),
    "flux_mole": (units.mol * units.s**-1 * units.m**-2),
    "flux_energy": (units.kg * units.s**-3),
    "velocity": (units.m * units.s**-1),
    "acceleration": (units.m * units.s**-2),
    "density_mass": (units.kg * units.m**-3),
    "density_mole": (units.mol * units.m**-3),
    "molecular_weight": (units.kg / units.mol),
    "energy": (units.kg * units.m**2 * units.s**-2),
    "energy_mass": (units.m**2 * units.s**-2),
    "energy_mole": (units.kg * units.m**2 * units.s**-2 * units.mol**-1),
    "dynamic_viscosity": (units.kg * units.m**-1 * units.s**-1),
    "entropy": (units.kg * units.m**2 * units.s**-2 * units.K**-1),
    "entropy_mass": (units.m**2 * units.s**-2 * units.K**-1),
    "entropy_mole": (units.kg * units.m**2 * units.s**-2 * units.K**-1 * units.mol**-1),
    "power": (units.kg * units.m**2 * units.s**-3),
    "voltage": (units.kg * units.m**2 * units.s**-3 * units.ampere**-1),
    "pressure": (units.kg * units.m**-1 * units.s**-2),
    "heat_capacity_mass": (units.m**2 * units.s**-2 * units.K**-1),
    "heat_capacity_mole": (
        units.kg * units.m**2 * units.s**-2 * units.K**-1 * units.mol**-1
    ),
    "heat_transfer_coefficient": (units.kg * units.s**-3 * units.K**-1),
    "thermal_conductivity": (units.kg * units.m * units.s**-3 * units.K**-1),
    "gas_constant": (units.kg * units.m**2 * units.s**-2 * units.K**-1 * units.mol**-1),
}


@pytest.mark.unit
@pytest.mark.parametrize("quantity", derived_quantities.keys())
def test_derived_units(unit_set, quantity):
    assert_units_equivalent(unit_set[quantity], derived_quantities[quantity])


@pytest.mark.unit
def test_add_default_units_extra_arg():
    m = ConcreteModel()

    m.meta_object = PropertyClassMetadata()

    with pytest.raises(
        TypeError,
        match="Unexpected argument for base quantities found when creating "
        "UnitSet. Please ensure that units are only defined for the seven "
        "base quantities.",
    ):
        m.meta_object.add_default_units(
            {"foo": "bar"},
        )


@pytest.mark.unit
def test_add_properties():
    m = ConcreteModel()
    m.meta_object = PropertyClassMetadata()

    m.meta_object.add_properties(
        {
            "flow_vol": {"method": "foo", "supported": True},
            "conc_mol_comp": {"method": "bar", "supported": True},
            "mw_phase": {"required": True, "supported": False},
            "dens_mol_phase_comp": {
                "method": "baz",
                "supported": True,
                "valid_range": (1, 10),
            },
        }
    )

    assert m.meta_object._properties.flow_vol._none.method == "foo"
    assert m.meta_object._properties.flow_vol._none.supported
    assert not m.meta_object._properties.flow_vol._none.required
    assert m.meta_object._properties.flow_vol._none.valid_range is None

    assert m.meta_object._properties.flow_vol._comp.method is None
    assert not m.meta_object._properties.flow_vol._comp.supported
    assert not m.meta_object._properties.flow_vol._comp.required
    assert m.meta_object._properties.flow_vol._comp.valid_range is None

    assert m.meta_object._properties.flow_vol._phase.method is None
    assert not m.meta_object._properties.flow_vol._phase.supported
    assert not m.meta_object._properties.flow_vol._phase.required
    assert m.meta_object._properties.flow_vol._phase.valid_range is None

    assert m.meta_object._properties.flow_vol._phase_comp.method is None
    assert not m.meta_object._properties.flow_vol._phase_comp.supported
    assert not m.meta_object._properties.flow_vol._phase_comp.required
    assert m.meta_object._properties.flow_vol._phase_comp.valid_range is None

    assert m.meta_object._properties.conc_mol._none.method is None
    assert not m.meta_object._properties.conc_mol._none.supported
    assert not m.meta_object._properties.conc_mol._none.required
    assert m.meta_object._properties.conc_mol._none.valid_range is None

    assert m.meta_object._properties.conc_mol._comp.method == "bar"
    assert m.meta_object._properties.conc_mol._comp.supported
    assert not m.meta_object._properties.conc_mol._comp.required
    assert m.meta_object._properties.conc_mol._comp.valid_range is None

    assert m.meta_object._properties.conc_mol._phase.method is None
    assert not m.meta_object._properties.conc_mol._phase.supported
    assert not m.meta_object._properties.conc_mol._phase.required
    assert m.meta_object._properties.conc_mol._phase.valid_range is None

    assert m.meta_object._properties.conc_mol._phase_comp.method is None
    assert not m.meta_object._properties.conc_mol._phase_comp.supported
    assert not m.meta_object._properties.conc_mol._phase_comp.required
    assert m.meta_object._properties.conc_mol._phase_comp.valid_range is None

    assert m.meta_object._properties.mw._none.method is None
    assert not m.meta_object._properties.mw._none.supported
    assert not m.meta_object._properties.mw._none.required
    assert m.meta_object._properties.mw._none.valid_range is None

    assert m.meta_object._properties.mw._comp.method is None
    assert not m.meta_object._properties.mw._comp.supported
    assert not m.meta_object._properties.mw._comp.required
    assert m.meta_object._properties.mw._comp.valid_range is None

    assert m.meta_object._properties.mw._phase.method is None
    assert not m.meta_object._properties.mw._phase.supported
    assert m.meta_object._properties.mw._phase.required
    assert m.meta_object._properties.mw._phase.valid_range is None

    assert m.meta_object._properties.mw._phase_comp.method is None
    assert not m.meta_object._properties.mw._phase_comp.supported
    assert not m.meta_object._properties.mw._phase_comp.required
    assert m.meta_object._properties.mw._phase_comp.valid_range is None

    assert m.meta_object._properties.dens_mol._none.method is None
    assert not m.meta_object._properties.dens_mol._none.supported
    assert not m.meta_object._properties.dens_mol._none.required
    assert m.meta_object._properties.dens_mol._none.valid_range is None

    assert m.meta_object._properties.dens_mol._comp.method is None
    assert not m.meta_object._properties.dens_mol._comp.supported
    assert not m.meta_object._properties.dens_mol._comp.required
    assert m.meta_object._properties.dens_mol._comp.valid_range is None

    assert m.meta_object._properties.dens_mol._phase.method is None
    assert not m.meta_object._properties.dens_mol._phase.supported
    assert not m.meta_object._properties.dens_mol._phase.required
    assert m.meta_object._properties.dens_mol._phase.valid_range is None

    assert m.meta_object._properties.dens_mol._phase_comp.method == "baz"
    assert m.meta_object._properties.dens_mol._phase_comp.supported
    assert not m.meta_object._properties.dens_mol._phase_comp.required
    assert m.meta_object._properties.dens_mol._phase_comp.valid_range == (1, 10)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "John Eslick"

from .phi_ideal_type01 import phi_ideal_expressions_type01
from .phi_ideal_type02 import phi_ideal_expressions_type02
from .phi_ideal_type03 import phi_ideal_expressions_type03
from .phi_residual_type01 import phi_residual_expressions_type01
from .phi_residual_type02 import phi_residual_expressions_type02
from .phi_residual_type03 import phi_residual_expressions_type03
from .phi_residual_type04 import phi_residual_expressions_type04
from .surface_tension_type01 import surface_tension_type01
from .sat_delta_approx import sat_delta_type01, sat_delta_type02

phi_residual_types = {
    0: None,  # custom
    1: phi_residual_expressions_type01,
    2: phi_residual_expressions_type02,
    3: phi_residual_expressions_type03,
    4: phi_residual_expressions_type04,
}

phi_ideal_types = {
    0: None,  # custom
    1: phi_ideal_expressions_type01,
    2: phi_ideal_expressions_type02,
    3: phi_ideal_expressions_type03,
}

delta_sat_types = {
    0: None,  # custom
    1: sat_delta_type01,
    2: sat_delta_type02,
}

surface_tension_types = {
    0: None,
    1: surface_tension_type01,
}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for ControlVolumeBlockData.

Author: Andrew Lee
"""
import pytest

from pyomo.environ import check_optimal_termination, ConcreteModel, value, Var, units
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from idaes.models.unit_models.plug_flow_reactor import PFR

from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)
from idaes.models.properties.examples.saponification_reactions import (
    SaponificationReactionParameterBlock,
)
from idaes.core.util.model_statistics import (
    number_variables,
    number_total_constraints,
    number_unused_variables,
    number_derivative_variables,
)
from idaes.core.util.testing import (
    PhysicalParameterTestBlock,
    ReactionParameterTestBlock,
    initialization_tester,
)
from idaes.core.solvers import get_solver
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    SingleControlVolumeUnitInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
@pytest.mark.component
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.reactions = ReactionParameterTestBlock(property_package=m.fs.properties)

    m.fs.unit = PFR(property_package=m.fs.properties, reaction_package=m.fs.reactions)

    # Check unit config arguments
    assert len(m.fs.unit.config) == 19

    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_heat_transfer
    assert not m.fs.unit.config.has_pressure_change
    assert not m.fs.unit.config.has_equilibrium_reactions
    assert not m.fs.unit.config.has_phase_equilibrium
    assert not m.fs.unit.config.has_heat_of_reaction
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.reaction_package is m.fs.reactions

    assert m.fs.unit.config.length_domain_set == [0.0, 1.0]
    assert m.fs.unit.config.transformation_method == "dae.finite_difference"
    assert m.fs.unit.config.transformation_scheme == "BACKWARD"
    assert m.fs.unit.config.finite_elements == 20
    assert m.fs.unit.config.collocation_points == 3


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(
            property_package=m.fs.properties
        )

        m.fs.unit = PFR(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            has_equilibrium_reactions=False,
            has_heat_transfer=True,
            has_heat_of_reaction=True,
            has_pressure_change=True,
        )

        m.fs.unit.inlet.flow_vol.fix(1.0)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(1e-8)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(1e-8)

        m.fs.unit.inlet.temperature.fix(303.15)
        m.fs.unit.inlet.pressure.fix(101325.0)

        m.fs.unit.length.fix(0.5)
        m.fs.unit.area.fix(0.1)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, sapon):
        assert hasattr(sapon.fs.unit, "inlet")
        assert len(sapon.fs.unit.inlet.vars) == 4
        assert hasattr(sapon.fs.unit.inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet, "temperature")
        assert hasattr(sapon.fs.unit.inlet, "pressure")

        assert hasattr(sapon.fs.unit, "outlet")
        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert isinstance(sapon.fs.unit.area, Var)
        assert isinstance(sapon.fs.unit.length, Var)
        assert isinstance(sapon.fs.unit.volume, Var)
        assert hasattr(sapon.fs.unit, "performance_eqn")
        assert hasattr(sapon.fs.unit.control_volume, "heat")
        assert hasattr(sapon.fs.unit, "heat_duty")
        assert hasattr(sapon.fs.unit, "deltaP")

        assert number_variables(sapon) == 654
        assert number_total_constraints(sapon) == 590
        assert number_unused_variables(sapon) == 14
        assert number_derivative_variables(sapon) == 0

    @pytest.mark.component
    def test_structural_issues(self, sapon):
        dt = DiagnosticsToolbox(sapon)
        dt.assert_no_structural_warnings()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, sapon):
        initialization_tester(sapon)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, sapon):
        assert (
            pytest.approx(101325.0, abs=1e-2) == sapon.fs.unit.outlet.pressure[0].value
        )
        assert (
            pytest.approx(303.6, abs=1e-2) == sapon.fs.unit.outlet.temperature[0].value
        )
        assert pytest.approx(62.29, abs=1e-2) == value(
            sapon.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, sapon):
        assert (
            abs(
                value(
                    sapon.fs.unit.inlet.flow_vol[0] - sapon.fs.unit.outlet.flow_vol[0]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    sapon.fs.unit.inlet.flow_vol[0]
                    * sum(
                        sapon.fs.unit.inlet.conc_mol_comp[0, j]
                        for j in sapon.fs.properties.component_list
                    )
                    - sapon.fs.unit.outlet.flow_vol[0]
                    * sum(
                        sapon.fs.unit.outlet.conc_mol_comp[0, j]
                        for j in sapon.fs.properties.component_list
                    )
                )
            )
            <= 1e-6
        )

        hrxn = 0
        for x in sapon.fs.unit.control_volume.length_domain:
            if x != 0:
                hrxn += value(
                    sapon.fs.unit.control_volume.heat_of_reaction[0, x]
                    * (x - sapon.fs.unit.control_volume.length_domain.prev(x))
                    * sapon.fs.unit.control_volume.length
                )
        assert pytest.approx(1847000, abs=1e3) == hrxn
        assert (
            abs(
                value(
                    (
                        sapon.fs.unit.inlet.flow_vol[0]
                        * sapon.fs.properties.dens_mol
                        * sapon.fs.properties.cp_mol
                        * (
                            sapon.fs.unit.inlet.temperature[0]
                            - sapon.fs.properties.temperature_ref
                        )
                    )
                    - (
                        sapon.fs.unit.outlet.flow_vol[0]
                        * sapon.fs.properties.dens_mol
                        * sapon.fs.properties.cp_mol
                        * (
                            sapon.fs.unit.outlet.temperature[0]
                            - sapon.fs.properties.temperature_ref
                        )
                    )
                    + hrxn
                )
            )
            <= 1e-6
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, sapon):
        dt = DiagnosticsToolbox(sapon)
        dt.assert_no_numerical_warnings()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, sapon):
        perf_dict = sapon.fs.unit._get_performance_contents()

        assert perf_dict == {"vars": {"Area": sapon.fs.unit.area}}


class TestInitializers:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(
            property_package=m.fs.properties
        )

        m.fs.unit = PFR(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            has_equilibrium_reactions=False,
            has_heat_transfer=True,
            has_heat_of_reaction=True,
            has_pressure_change=True,
        )

        m.fs.unit.inlet.flow_vol[0].set_value(1.0)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].set_value(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].set_value(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].set_value(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].set_value(0.0)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].set_value(0.0)

        m.fs.unit.inlet.temperature[0].set_value(303.15)
        m.fs.unit.inlet.pressure[0].set_value(101325.0)

        m.fs.unit.length.fix(0.5)
        m.fs.unit.area.fix(0.1)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.component
    def test_general_hierarchical(self, model):
        initializer = SingleControlVolumeUnitInitializer()
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert (
            pytest.approx(101325.0, rel=1e-5) == model.fs.unit.outlet.pressure[0].value
        )
        assert (
            pytest.approx(303.593, rel=1e-5)
            == model.fs.unit.outlet.temperature[0].value
        )
        assert pytest.approx(62.29202, rel=1e-5) == value(
            model.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]
        )

        assert not model.fs.unit.inlet.flow_vol[0].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "H2O"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fixed

        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed

    @pytest.mark.component
    def test_block_triangularization(self, model):
        initializer = BlockTriangularizationInitializer(constraint_tolerance=1e-5)

        # Need to exclude unused variables for x=0 in post-check
        initializer.initialize(model.fs.unit, exclude_unused_vars=True)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert (
            pytest.approx(101325.0, rel=1e-5) == model.fs.unit.outlet.pressure[0].value
        )
        assert (
            pytest.approx(303.593, rel=1e-5)
            == model.fs.unit.outlet.temperature[0].value
        )
        assert pytest.approx(62.29202, rel=1e-5) == value(
            model.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]
        )

        assert not model.fs.unit.inlet.flow_vol[0].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "H2O"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fixed

        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest
import base64
import io
import os
import tarfile
import tempfile
import shutil
from platform import machine

from pyomo.common.download import FileDownloader

import idaes
import idaes.commands.util.download_bin as dlb
import idaes.logger as idaeslog
import idaes.config as idaes_config


_log = idaeslog.getLogger(__name__)


def _del_data_file(path):
    try:
        os.remove(path)
    except OSError:
        pass


@pytest.mark.unit
def test_get_file_downloader():
    fd, arch = dlb._get_file_downloader(insecure=True, cacert=None)

    assert isinstance(fd, FileDownloader)
    assert fd.insecure
    assert fd.cacert is None


@pytest.mark.unit
def test_get_file_downloader_cacert_error():
    with pytest.raises(
        RuntimeError, match="cacert='foo' does not refer to a valid file."
    ):
        dlb._get_file_downloader(insecure=False, cacert="foo")


@pytest.mark.unit
def test_get_arch_and_platform_auto():
    fd = FileDownloader()

    arch, platform = dlb._get_arch_and_platform(fd, platform="auto")

    assert arch == fd.get_sysinfo()

    if arch[0] == "linux":
        assert platform == fd.get_os_version().replace(".", "")
    else:
        assert platform == arch[0]


@pytest.mark.unit
def test_get_release_platform_mapping():
    mach = idaes.config.canonical_arch(machine())
    for p, m in idaes_config.binary_distro_map.items():
        if mach == "aarch64" and m == "el7":
            continue
        if mach == "x86_64" and m == "darwin":
            continue
        output = dlb._get_release_platform(p)
        assert output.startswith(m)
        assert output in idaes_config.base_platforms


@pytest.mark.unit
def test_get_release_url_release():
    output = dlb._get_release_url("foo/", None)

    assert output == "https://github.com/IDAES/idaes-ext/releases/download/foo"


@pytest.mark.unit
def test_get_release_url_release_and_url():
    output = dlb._get_release_url("foo/", "bar/")

    assert output == "https://github.com/IDAES/idaes-ext/releases/download/foo"


@pytest.mark.unit
def test_get_release_url():
    output = dlb._get_release_url(None, "bar/")

    assert output == "bar"


@pytest.mark.unit
def test_get_release_url_none():
    with pytest.raises(Exception, match="Must provide a location to download binaries"):
        dlb._get_release_url(None, None)


@pytest.mark.unit
def test_get_checksum_paths():
    cto, cfrom = dlb._get_checksum_paths("foo", "bar")

    assert cto == os.path.join("foo", f"sha256sum_bar.txt")
    assert (
        cfrom
        == "https://raw.githubusercontent.com/IDAES/idaes-ext/main/releases/sha256sum_bar.txt"
    )


@pytest.mark.unit
def test_download_checksum():
    # Mock paths to avoid downloads
    fd = FileDownloader()
    fd.retrieve_url = lambda url: bytes("\n", encoding="utf-8")

    tmpdir = tempfile.mkdtemp()
    target = os.path.join(tmpdir, "bin.txt")

    output = dlb._download_checksum(fd, target, None)

    assert output is None

    assert os.path.getsize(target) == 1

    shutil.rmtree(tmpdir)


@pytest.mark.unit
def test_read_checksum():
    path = os.path.dirname(__file__)
    testfile = os.path.join(path, "checksum.txt")

    checksum = dlb._read_checksum_file(testfile)

    assert checksum == {
        "idaes-lib-centos6-64.tar.gz": "449457bc91fac41e8c443b4d9cf2ba9181201e84318e890833f9685c01dcd531",
        "idaes-solvers-centos6-64.tar.gz": "2efda2ce61f9fc1d7180b5b141077ce52c2682f6532e9c8d873164af13162948",
        "idaes-lib-centos7-64.tar.gz": "dfae8a6e1bb4afe27220062d8b47abd780207f7fd85cf27126cf2e2c41ef6d36",
        "idaes-solvers-centos7-64.tar.gz": "ae465b064b7e74c91127558c6f0ebbae1b01756e5c4d59d5681a70088a70ffcd",
        "idaes-lib-centos8-64.tar.gz": "5301a517e098262be2e4536207114c855eb9c0cf17aad085a4964ba64674e457",
        "idaes-solvers-centos8-64.tar.gz": "a597b72d71b12e019a0d629e5d8eb089382d8051488f6a51f597511fc2872b3f",
        "idaes-lib-ubuntu1804-64.tar.gz": "bc6bbf8e2ede3e5afc80fed8656eddb575a2464cebb593afdab9171c2b5723e6",
        "idaes-solvers-ubuntu1804-64.tar.gz": "25299b15e270a2e7224e5aa4d6665bec074ceb727d7a4cf87b11d01fc96acdef",
        "idaes-lib-ubuntu1910-64.tar.gz": "6fbc3034d8c94763c8db9798d6349eb31a1980af92829cef5fecdfca3bc0454a",
        "idaes-solvers-ubuntu1910-64.tar.gz": "f635ceccb71c0e5c6c0b8bcb26b507d5531ec34630aa897d794d0e7d3dc0d401",
        "idaes-lib-ubuntu2004-64.tar.gz": "272ee8fc338074e056fed2c6115cc504526c389e5880c8819af7e6c7b9f1fc61",
        "idaes-solvers-ubuntu2004-64.tar.gz": "586fe646436e27f3e3e0f4c7a3c983c500f5048f2e65eb28f9303d790d600986",
        "idaes-lib-windows-64.tar.gz": "3e17985dfc12c7dea8da47b6d15c45d2d27e68accaacc521d708ef31c08d001e",
        "idaes-solvers-windows-64.tar.gz": "c9c141a38b208025bd16a3f3af8d6deb2dcd03777b63c04c5f8842f2743b7e7b",
    }


@pytest.mark.parametrize("platform", idaes_config.base_platforms)
@pytest.mark.unit
def test_create_download_package(platform):
    pname, ptar, ftar, furl = dlb._create_download_package(
        platform, "foo", "bar", [], False, False
    )

    assert len(ptar) == 2
    assert (os.path.join("foo", f"idaes-lib-{platform}.tar.gz")) in ptar
    assert (os.path.join("foo", f"idaes-solvers-{platform}.tar.gz")) in ptar

    assert ftar == [f"idaes-lib-{platform}.tar.gz", f"idaes-solvers-{platform}.tar.gz"]
    assert furl == [
        f"bar/idaes-lib-{platform}.tar.gz",
        f"bar/idaes-solvers-{platform}.tar.gz",
    ]


@pytest.mark.parametrize("platform", idaes_config.base_platforms)
@pytest.mark.unit
def test_create_download_package_extras(platform):
    pname, ptar, ftar, furl = dlb._create_download_package(
        platform, "foo", "bar", ["petsc", "baz"], False, False
    )

    # baz should be ignored as an unknown extra
    assert len(ptar) == 3
    assert (os.path.join("foo", f"idaes-lib-{platform}.tar.gz")) in ptar
    assert (os.path.join("foo", f"idaes-solvers-{platform}.tar.gz")) in ptar
    assert (os.path.join("foo", f"idaes-petsc-{platform}.tar.gz")) in ptar

    assert ftar == [
        f"idaes-petsc-{platform}.tar.gz",
        f"idaes-lib-{platform}.tar.gz",
        f"idaes-solvers-{platform}.tar.gz",
    ]
    assert furl == [
        f"bar/idaes-petsc-{platform}.tar.gz",
        f"bar/idaes-lib-{platform}.tar.gz",
        f"bar/idaes-solvers-{platform}.tar.gz",
    ]


@pytest.mark.parametrize("platform", idaes_config.base_platforms)
@pytest.mark.unit
def test_create_download_package_extras_only(platform):
    pname, ptar, ftar, furl = dlb._create_download_package(
        platform, "foo", "bar", ["petsc", "baz"], True, False
    )

    # baz should be ignored as an unknown extra
    assert len(ptar) == 1
    assert (os.path.join("foo", f"idaes-petsc-{platform}.tar.gz")) in ptar

    assert ftar == [f"idaes-petsc-{platform}.tar.gz"]
    assert furl == [f"bar/idaes-petsc-{platform}.tar.gz"]


@pytest.mark.parametrize("platform", idaes_config.base_platforms)
@pytest.mark.unit
def test_create_download_package_library_only(platform):
    pname, ptar, ftar, furl = dlb._create_download_package(
        platform, "foo", "bar", [], False, True
    )

    assert len(ptar) == 1
    assert (os.path.join("foo", f"idaes-lib-{platform}.tar.gz")) in ptar

    assert ftar == [f"idaes-lib-{platform}.tar.gz"]
    assert furl == [f"bar/idaes-lib-{platform}.tar.gz"]


@pytest.mark.unit
def test_download_package():
    # Mock paths to avoid downloads
    fd = FileDownloader()
    fd.retrieve_url = lambda url: bytes("\n", encoding="utf-8")

    tmpdir = tempfile.mkdtemp()
    target = os.path.join(tmpdir, "bin.txt")

    output = dlb._download_package(fd, "foo", None, target, "bar")

    assert output is None

    assert os.path.getsize(target) == 1

    shutil.rmtree(tmpdir)


@pytest.mark.unit
def test_dl_bin_unknown():
    _del_data_file(os.path.join(idaes.testing_directory, "version_lib.txt"))
    _del_data_file(os.path.join(idaes.testing_directory, "version_solvers.txt"))
    with pytest.raises(Exception):
        dlb.download_binaries(
            platform="unknown platform",
            release=idaes.config.default_binary_release,
            to_path="testing",
        )


@pytest.mark.unit
def test_verify_tarfile_members():
    # Tarball with:
    #
    # [DIRTYPE] dir1
    # [SYMTYPE] dir1/dir2 -> ..
    # [LNKTYPE] dir1/dir2 -> dir1/dir2
    # [SYMTYPE] dir1/dir2/dir3 -> ..
    # [REGTYPE] dir1/dir2/dir3/file
    #
    softlink_tar_gz = (
        b"H4sICLbv92MAA3NvZnRsaW5rLnRhcgDt0+0KgjAUgOFdyq5Ad7a5XU9gwSIItO4/l0EFR"
        b"QlO+nifHwoeYQdfbFMntSrLDGJslBEfw+39Sol3sclPox/mEo1Vuim819mxP6w6rdW27V"
        b"Pq9k/fezX/Um3uP1xswTPG/nFCfxskKG2rquBWF/T/wP5ejCgtSyxH//ET54src8aU/sG"
        b"4Ye58Y/j/l3Dfv96k3Xr2M3LgEPx7/UXy3FqntJl9kwf+vD8AAAAAAAAAAAAA4DecAEnw"
        b"Tw0AKAAA"
    )
    data = io.BytesIO(base64.b64decode(softlink_tar_gz))
    tar = tarfile.open(None, "r:gz", data)
    # verify tarfile contents:
    # for f in tar.getmembers():
    #     print(f.type, f.name, f.linkname)
    with pytest.raises(
        Exception,
        match="Tarball None contained potentially unsafe member dir1/dir2/dir3/file",
    ):
        dlb._verify_tar_member_targets(tar, os.path.abspath(""))

    # Tarball with:
    #
    # [DIRTYPE] dir1
    # [SYMTYPE] dir1/dir2 -> ..
    # [LNKTYPE] dir1/dir2 -> dir1/dir2
    # [LNKTYPE] dir1/dir2/dir3 -> dir1/dir2
    # [REGTYPE] dir3/file
    #
    hardlink_tar_gz = (
        b"H4sICHg++GMAA2hhcmRsaW5rLnRhcgDt1FkKwjAUQNG3lKzA5qUZ1iNUISIIre7fxor6I"
        b"w6Q4nDPRwpNIY9caJd7baQuO0opiFWf4u3zStS3KZS3yY/7mqwTEyrPdXIY9sveGNl0Q8"
        b"797u53j/a/VFf6j4ureMbUP73Q30WNYtxiUXGqM/p/YH+vVsXoHMPRf7risrR1znijf/K"
        b"O/nMo1Zt13q4qnlECx+if6h9Vy/8/OC/GVpzp4s/7AwAAAAAAAAAAAAB+wxHqgNJYACgA"
        b"AA=="
    )
    data = io.BytesIO(base64.b64decode(hardlink_tar_gz))
    tar = tarfile.open(None, "r:gz", data)
    # verify tarfile contents:
    # for f in tar.getmembers():
    #     print(f.type, f.name, f.linkname)
    with pytest.raises(
        Exception, match="Tarball None contained potentially unsafe member dir3/file"
    ):
        dlb._verify_tar_member_targets(tar, os.path.abspath(""))
